import React, {useCallback, useContext} from "react";
import {Smartable, useTable} from "./Smartable";

/**
 * Basic column key type.
 */
export type ColumnKey = string|symbol;

/**
 * Smartable columns definition.
 */
export type Columns<K extends ColumnKey> = Record<K, Column>;

/**
 * Define columns for a Smartable.
 * @param columns
 */
export function createColumns<K extends ColumnKey>(...columns: [K, Column][]): Columns<K>
{
	return Object.fromEntries(
		columns,
	) as Columns<K>;
}

/**
 * Smartable column definition.
 */
export interface Column<T = any>
{
	/**
	 * Column title element.
	 */
	title: React.ReactNode;

	/**
	 * Column cell default element.
	 */
	cellElement?: React.ReactElement;

	/**
	 * Cells data comparison in the column.
	 * @param a First data to compare.
	 * @param b Second data to compare.
	 */
	compare?: (a: T, b: T) => number;
}

/**
 * Create a column for a Smartable.
 * @param key The column key, which should be unique per Smartable instance.
 * @param column The column definition.
 */
export function createColumn<K extends ColumnKey>(key: K, column: Column): [K, Column]
{
	return [key, column];
}

/**
 * Column sort type.
 */
export enum SortType
{
	ASC = "asc",
	DESC = "desc",
}

/**
 * Column sort state.
 */
export interface SortState
{
	/**
	 * Sort type (ascending or descending).
	 */
	type: SortType;

	/**
	 * Sort order.
	 */
	order: number;
}

/**
 * Table column context data.
 */
export interface ColumnContextData<CK extends ColumnKey>
{
	/**
	 * Column key.
	 */
	key: ColumnKey;

	/**
	 * Column definition.
	 */
	column: Column;

	/**
	 * Column sort state.
	 */
	sortState?: SortState;
}

export const ColumnContext = React.createContext<ColumnContextData<ColumnKey>>(undefined);

/**
 * Hook to get current column data.
 */
export function useColumn<CK extends ColumnKey>(smartable?: Smartable<CK>): ColumnContextData<CK>
{
	return useContext(ColumnContext);
}

/**
 * Default column heading component.
 */
export function ColumnHeading()
{
	// Get current column data.
	const {key, column, sortState} = useColumn();

	// Get column sort state setter.
	const {setColumnSortState} = useTable();

	// Initialize handle click function.
	const handleClick = useCallback((event: React.MouseEvent) => {
		if (event.button == 0)
		{ // Normal click (usually left click).
			// Toggle sort type.
			setColumnSortState(key, sortState?.type == SortType.ASC ? SortType.DESC : SortType.ASC);
		}
		else if (event.button == 2 || event.button == 1)
		{ // Alt click (usually right or middle click).
			// Reset sort type.
			setColumnSortState(key, null);
		}
	}, [key, sortState, setColumnSortState]);

	// Disable context menu function.
	const disableContextMenu = useCallback((event: React.MouseEvent) => {
		event.preventDefault();
		return false;
	}, []);

	return (
		<th className={sortState?.type ?? undefined} data-sort-order={sortState?.order ?? undefined}
		    onMouseDown={handleClick} onContextMenu={disableContextMenu}>
			{column.title}

			{sortState?.order && <span className={"order"}>{sortState.order}</span>}
		</th>
	);
}

/**
 * Auto column context provider from table context.
 */
export function AutoColumnContextProvider({columnKey, children}: React.PropsWithChildren<{columnKey: ColumnKey;}>)
{
	// Get table data.
	const table = useTable();

	return (
		<ColumnContext.Provider value={{
			key: columnKey,
			// Get current column data from table data.
			column: table.columns[columnKey],
			// Get current column sort state from table data.
			sortState: table.columnsSortState?.[columnKey],
		}}>
			{children}
		</ColumnContext.Provider>
	)
}
